﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Reversi
{
    // AI 3号（探索：アルファベータ法、評価：石の位置による重みづけ）
    class AI3 : AI
    {
        public AI3(int depth, int order) : base(depth, order){}

        // 評価の重みづけ
        private int[,] WEIGHT = {
            {+50, -10, +5, +5, +5, +5, -10, +50},
            {-10, -10, -2, -2, -2, -2, -10, -10},
            { +5,  -2, +2, +1, +1, +2,  -2,  +5},
            { +5,  -2, +1, +1, +1, +1,  -2,  +5},
            { +5,  -2, +1, +1, +1, +1,  -2,  +5},
            { +5,  -2, +2, +1, +1, +2,  -2,  +5},
            {-10, -10, -2, -2, -2, -2, -10, -10},
            {+50, -10, +5, +5, +5, +5, -10, +50}
        };

        // 評価関数
        private int evalFunc(Board board)
        {
            int eval = 0;

            for (int x = 0; x < Board.SIZE; x++)
            {
                for (int y = 0; y < Board.SIZE; y++)
                {
                    int owner = board.getStone(x, y);
                    if (owner == myOrder)
                    {
                        eval += WEIGHT[x, y];
                    }
                    else if (owner == -myOrder)
                    {
                        eval -= WEIGHT[x, y];
                    }
                }
            }
            return eval;
        }

        // 次の手を取得する
        public override Move getNextMove(Board board)
        {
            // 次の手を考える
            readAlphaBeta(board, 0, int.MinValue, int.MaxValue);

            return bestMove;
        }

        // アルファベータ法による先読み (再帰)
        public int readAlphaBeta(Board board, int depth, int alpha, int beta)
        {
            int turnHolder = board.turnHolder;

            // 先読み深さの末端に到達したら評価値を返す
            if (depth == maxDepth)
            {
                // 評価関数
                int eval = evalFunc(board);

                return eval;
            }

            // 次の局面を列挙
            ulong nextMoves = board.enumNextMoves();

            // 自分の手番なら最も自分に有利な手を選択（自分にとっての最善手）
            // 相手の手番なら最も自分に不利な手を選択（相手にとっての最善手）
            if (nextMoves == 0)
            {
                // パスの場合
                Board nextBoard = board;
                nextBoard.doMove(Move.PASS);
 
                // 再帰呼び出し
                int eval = readAlphaBeta(nextBoard, depth + 1, alpha, beta);
                return eval;
            }
            else for (int i = 0; i < Board.SIZE * Board.SIZE; i++)
            {
                if ((nextMoves & (1ul << i)) != 0)
                {
                    int x = i % Board.SIZE;
                    int y = i / Board.SIZE;
                    Move move = new Move(x,y);
                    Board nextBoard = board;
                    nextBoard.doMove(move);

                    // 再帰呼び出し
                    int eval = readAlphaBeta(nextBoard, depth + 1, alpha, beta);

                    if ((turnHolder == myOrder) && (eval > alpha))
                    {
                        alpha = eval;
                        if (depth == 0)
                        {
                            bestMove = move;
                        }
                        if (alpha >= beta)
                        {
                            alpha = beta;
                            break; // カット
                        }
                    }
                    if ((turnHolder != myOrder) && (eval < beta))
                    {
                        beta = eval;
                        if (alpha >= beta)
                        {
                            beta = alpha;
                            break; // カット
                        }
                    }
                }
            }
            return (turnHolder == myOrder) ? alpha : beta;
        }
    }
}
